@@ -6,8 +6,8 @@
 # Author:  Vadim Zeitlin <vadim@wxwindows.org>
 # License: Public domain
 # Created: June 25, 2001
-# Version: 0.22
-# RCS-Id:  $Id: yahoo2mbox.pl,v 1.53 2006/06/11 12:13:48 zeitlin Exp $
+# Version: 0.24
+# RCS-Id:  $Id: yahoo2mbox.pl,v 1.56 2007-10-22 20:14:46 zeitlin Exp $
 ###############################################################################
 
 use strict;
@@ -93,10 +93,11 @@ sub ParseMessagePage($) {
         warn "$warnMsg.\n";
 
         if ( $debug ) {
-            open(BAD, ">bad.html") or die;
+            my $bad_file = "bad_${main::msg}.html";
+            open(BAD, ">$bad_file") or die "Can't make debug output $bad_file. $!\n";
             print BAD $html;
 
-            warn "Message text saved to file bad.html for debugging.\n"
+            warn "Message text saved to file $bad_file for debugging.\n"
         }
 
         return undef
@@ -195,31 +196,37 @@ sub MBOXify($) {
 
     # process the message line by line
     foreach my $line ( $inmsg =~ m/^.*$/ogm ) {
-        if ( $inHdr ) {
+        if ( !$inMsg ) {
+            # the first header line is supposed to be "From ", as usual
             if ( $line =~ /^From / ) {
                 $inMsg = 1;
-                # the token following "From " is supposed to be a single word,
-                # but Yahoo puts the full RFC 822 address there (e.g. something
-                # like "foo bar <foo@bar.com>"), massage it into the standard
-                # form
-                $line =~ s/From .*<([^>]+)>/From $1/;
             }
-            elsif ( !$inMsg ) {
-                # ignore everything until we see the "From ".
-                next;
+            else {
+                if ( $line !~ /^\s*$/ ) {
+                    warn "Ignoring unexpected text (\"$line\") before the message start."
+                }
+
+                # ignore everything before the beginning of the real message
+                next
             }
-            elsif ( !$line || ($line =~ /^\s*$/) ) {
+        }
+
+        if ( $inHdr ) {
+            if ( $line =~ /^\s*$/ ) {
                 # end of header
                 $inHdr = 0;
 
-                # fix empty line between headers and body so it actually is
+                # fix the line between headers and body to be really empty
                 $line = "";
 
-                # add our header at the end of the original one if needed
+                # add our header at the end of the original ones if needed
                 $msg .= "$x_yahoo_hdr_name: $main::msg\n" if $x_yahoo_hdr;
             }
-            elsif (( $line !~ /^[\w-]+: / ) && ( $line !~ /^\s+/ )) {
-                # looks like a header continuation
+            elsif ( $line !~ /^\s+/ &&
+                        $line !~ /^[\w-]+: / &&
+                            $line !~ /^From / ) {
+                # looks like a header continuation from which Yahoo ate the
+                # leading space -- restore it
                 $line = "  " . $line;
             }
         }
@@ -233,6 +240,9 @@ sub MBOXify($) {
         $msg .= $line . "\n";
     }
 
+    # messages are separated by blank lines in MBOX
+    $msg .= "\n";
+
     return $msg;
 }
 
@@ -260,7 +270,7 @@ sub DieConflictingOptions($$);
 # command line options parsing
 ###############################################################################
 
-my $VERSION_STRING = "0.22";
+my $VERSION_STRING = "0.24";
 
 # boolean: if true, suppress messages
 my $quiet = 0;
@@ -286,6 +296,12 @@ my $msgFirstGivenByUser = 1;
 # bool values mean usual thing, -1 means to resume if mailbox already exists
 my $resume = -1;
 
+# retry getting the messages we failed to retrieve the last time?
+my $retry   = 0;
+
+my @msgs   = ();
+my %msgSeen = ();
+
 # authentification info
 my ($username, $password);
 
@@ -321,6 +337,7 @@ the messages not already in the file are appended to it.
 --help          give the usage message showing the program options
 --version       show the program version and exit
 --verbose       give verbose informational messages (default)
+--debug         turns on debug output and saves messages script failed to parse
 --quiet         be silent, only error messages are given
 -o mbox         save the message to mbox instead of file named groupname
 --start=n       start retrieving messages at index n instead of 1
@@ -330,6 +347,7 @@ the messages not already in the file are appended to it.
 --resume        resumes previously interrupted downloading, needed if --start
                 is used, otherwise is automatic if mailbox already exists
 --noresume      don't resume, **overwrites** the existing output file if any
+--retry         resumes and tries to download the missing messages
 --user=name     login to eGroups using this username (default: guest login)
 --pass=pass     the password to use for login (default: none)
 --cookies=xxx   file to use to store cookies (default: none, 'netscape' uses
@@ -353,6 +371,7 @@ EOF
 # don't forget to modify the usage section above if you add more options!
 Getopt::Long::GetOptions
 (
+    'debug'         => sub { $debug = 1 },
     'verbose'       => sub { $quiet = 0 },
     'quiet'         => \$quiet,
     'version'       => \$version,
@@ -364,6 +383,7 @@ Getopt::Long::GetOptions
     'last=i'        => \$msgCountFromEnd,
     'noresume'      => sub { $resume = 0 },
     'resume'        => sub { $resume = 1 },
+    'retry'         => sub { $resume = 1; $retry = 1;},
     'user=s'        => \$username,
     'pass=s'        => \$password,
     'cookies=s'     => \$cookie_file,
@@ -532,6 +552,7 @@ if ( $resume && -r $mboxname && ($msgFirst == 0 || $resume == 1) ) {
             $msgFirst++;
         }
         elsif (/^$x_yahoo_hdr_name: (\d+)$/) {
+            $msgSeen{$1}++;
             # messages can be out of order in the mailbox (some mail clients do
             # this, apparently) so take the largest value, not the last one
             $numYahoo = $1 if $1 > $numYahoo;
@@ -542,7 +563,9 @@ if ( $resume && -r $mboxname && ($msgFirst == 0 || $resume == 1) ) {
         $msgFirst = $numYahoo + 1;
     }
 
-    print "Resuming at message $msgFirst\n" unless $quiet;
+    if ( !$retry ) {
+        print "Will resume at message $msgFirst\n" unless $quiet;
+    }
 }
 
 # we need to "login" to the service first: for this we need to get the cookie
@@ -666,51 +689,53 @@ while ( $response->is_redirect ) {
 }
 
 # determine the last message to retrieve (must be done after login)
-if ( !$msgLast ) {
-    print "Getting number of messages in group $groupname...\n" unless $quiet;
+print "Getting number of messages in group $groupname...\n" unless $quiet;
 
-    # determine the number of messages to retrieve
-    #
-    # TODO: we could get only the start of the document - would be faster
-    $request = GET $baseurl . $idxext;
-    $response = $ua->simple_request($request) ||
-       die "Failed to get the messages page for group $groupname.\n";
+# determine the number of messages to retrieve
+#
+# TODO: we could get only the start of the document - would be faster
+$request = GET $baseurl . $idxext;
+$response = $ua->simple_request($request) ||
+   die "Failed to get the messages page for group $groupname.\n";
 
-    # parse the header
-    my $messagesPage = $response->content;
-    my $headerParser = HTML::HeadParser->new;
+# parse the header
+my $messagesPage = $response->content;
+my $headerParser = HTML::HeadParser->new;
 
-    $headerParser->parse($messagesPage);
-    $_ = $headerParser->header('title') || "";
-    if ( !/\d+-(\d+) $lang{of} (\d+)/ ) {
-        # try to see why did it happen
-        if ( $messagesPage =~ /$lang{badgroup}/i ) {
-            die "The group $groupname doesn't seem to exist.\n";
-        }
-        elsif ( !$_ ) {
-            my $err = "Failed to access group page,";
-            if ( !$username ) {
-                die "$err you might need a valid username.\n";
-            }
-            else {
-                die "$err please check if you have specified valid password.\n";
-            }
+$headerParser->parse($messagesPage);
+$_ = $headerParser->header('title') || "";
+if ( !/\d+-(\d+) $lang{of} (\d+)/ ) {
+    # try to see why did it happen
+    if ( $messagesPage =~ /$lang{badgroup}/i ) {
+        die "The group $groupname doesn't seem to exist.\n";
+    }
+    elsif ( !$_ ) {
+        my $err = "Failed to access group page,";
+        if ( !$username ) {
+            die "$err you might need a valid username.\n";
         }
         else {
-            #open(PAGE, ">page.html") or die; print PAGE $messagesPage;
-
-            die "Unexpected title page format ($_).\n";
+            die "$err please check if you have specified valid password.\n";
         }
-    };
-
-    if ( $1 < $2 ) {
-        print "Warning: " . ($2 - $1) .
-              " messages in this group seem to be missing.\n" unless $quiet;
     }
+    else {
+        #open(PAGE, ">page.html") or die; print PAGE $messagesPage;
 
-    $msgLast = $1;
+        die "Unexpected title page format ($_).\n";
+    }
 };
 
+if ( $1 < $2 ) {
+    print "Warning: " . ($2 - $1) .
+          " messages in this group seem to be missing.\n" unless $quiet;
+}
+
+my $msgCount = $1;
+
+if ( !$msgLast ) {
+    $msgLast = $msgCount;
+}
+
 if ( $msgCountFromEnd && $msgCountFromEnd < $msgLast ) {
     $msgFirst = $msgLast - $msgCountFromEnd + 1;
 }
@@ -735,6 +760,13 @@ if ( $msgCountNext ) {
     $msgLast = $msgFirst + $msgCountNext - 1;
 }
 
+if ( $msgLast > $msgCount ) {
+    printf "Warning: index of the last message to retrieve ($msgLast) " .
+           "is greater than the number of messages in the group ($msgCount).\n"
+           unless $quiet;
+    $msgLast = $msgCount;
+}
+
 # append to the file, don't overwrite it
 open(OUTPUT, ">>$mboxname") || die "Failed to open $mboxname for writing: $!\n";
 
@@ -764,12 +796,40 @@ if ( $msgLast < $msgFirst ) {
 
 my $ok = 1;
 
-print "Retrieving messages $msgFirst..$msgLast: " unless $quiet;
+# retry to get the missing messages if requested
+if ( $retry ) {
+    my $missing = 0;
+    for ( my $j = 1; $j < $msgFirst; $j++ ) {
+       next if defined $msgSeen{$j};
+       $missing++;
+       push @msgs, $j;
+    }
+
+    if ( !$missing ) {
+        print "No missing messages to retrieve.\n" unless $quiet;
+        exit 0
+    }
+
+    print "Retrying to retrieve $missing missing messages: " unless $quiet;
+}
+else {
+    print "Retrieving messages $msgFirst..$msgLast: " unless $quiet;
+
+    # get all messages in order
+    for ( my $i = $msgFirst; $i <= $msgLast; $i++ ) {
+        push @msgs, $i;
+    }
+}
+
+warn "\n" if $debug;  # newline for debug output
 
 # the main loop over the messages
 MAINLOOP:
-for $main::msg ($msgFirst..$msgLast) {
+for $main::msg (@msgs) {
     my $msgUrl = $baseurl . sprintf($msgext, $main::msg);
+
+    warn "Downloading message $main::msg" if $debug;
+
     my $request = GET $msgUrl;
     my $response = $ua->simple_request($request);
 
@@ -866,7 +926,14 @@ sub DieConflictingOptions($$) {
 
 # History:
 #
-# 0.22 updated to deal with the new format of Yahoo pages
+# 0.24  Add --retry resume option to reget any missing messages (Max Baker)
+#       Save separate bad.html for each bad response to aid in debugging
+#       Added --debug as CLI option
+#
+# 0.23  handle messages beyond the last one correctly (Michael Kielsky)
+#       corrected missing blank lines between messages
+#
+# 0.22  updated to deal with the new format of Yahoo pages
 #
 # 0.21  deal correctly with --x-yahoo even if mailbox messages are out of order
 #       (Daniel Sutcliffe)
